/*
 *  IRNR.cpp
 *  voting
 *
 *  Created by Brian Olson on Mon Dec 01 2003.
 *  Copyright (c) 2003 Brian Olson. All rights reserved.
 *
 */

#include "IRNR.h"
#include "Voter.h"
#include <stdio.h>
#include <math.h>

#ifndef TRACK_MAX
#define TRACK_MAX 0
#endif
#ifndef TRACK_DQ
#define TRACK_DQ 0
#endif

IRNR::IRNR( char* nameIn )
: VotingSystem( nameIn )
{
}
void IRNR::runElection( int* winnerR, const VoterArray& they ) {
    float* counterweights;
    bool* active;
    double* talley;
    int numActive;
    int c, v;
    double min;
    int minc;
#if TRACK_MAX
    double max;
    int maxc, oldmaxc = -1;
#endif

    counterweights = new float[they.numv];
    active = new bool[they.numc];
    talley = new double[they.numc];
    for ( c = 0; c < they.numc; c++ ) {
	active[c] = true;
	talley[c] = 0.0;
    }
    numActive = they.numc;
    for ( v = 0; v < they.numv; v++ ) {
	double ts;
	ts = 0.0;
	for ( c = 0; c < they.numc; c++ ) {
	    ts += fabs( they[v].getPref( c ) );
	}
	counterweights[v] = 1.0 / ts;
	for ( c = 0; c < they.numc; c++ ) {
	    talley[c] += they[v].getPref( c ) * counterweights[v];
	}
    }
    do {
	minc = -1;
	min = they.numv * 2.0;
#if TRACK_MAX
	maxc = -1;
	max = they.numv * -2.0;
#endif
	for ( c = 0; c < they.numc; c++ ) if ( active[c] ) {
	    if ( talley[c] < min ) {
		min = talley[c];
		minc = c;
	    }
#if TRACK_MAX
	    if ( talley[c] > max ) {
		max = talley[c];
		maxc = c;
	    }
#endif	
	}
#if TRACK_MAX
	if ( maxc != oldmaxc ) {
	    printf( "IRNR: maxc changed from %d to %d\n", oldmaxc, maxc );
	    oldmaxc = maxc;
	}
#endif
	if ( minc == -1 ) {
	    fprintf( stderr, "%s:%d couldn't find a loser, min %f\n", __FILE__, __LINE__, min );
	    for ( c = 0; c < they.numc; c++ ) {
		fprintf( stderr, "\ttalley[%d] = %f (%s)\n", c, talley[c], active[c] ? "ON " : "off" );
	    }
	    break;
	}
	active[minc] = false;
#if TRACK_DQ
	printf( "IRNR: %d disqualified\n", minc );
#endif
	numActive--;
	if ( numActive == 1 ) break;
	for ( c = 0; c < they.numc; c++ ) {
	    talley[c] = 0.0;
	}
	for ( v = 0; v < they.numv; v++ ) {
	    double ts;
	    ts = 0.0;
	    for ( c = 0; c < they.numc; c++ ) if ( active[c] ) {
		ts += fabs( they[v].getPref( c ) );
	    }
	    counterweights[v] = 1.0 / ts;
	    for ( c = 0; c < they.numc; c++ ) if ( active[c] ) {
		talley[c] += they[v].getPref( c ) * counterweights[v];
	    }
	}
    } while ( true );
    for ( c = 0; c < they.numc; c++ ) if ( active[c] ) {
	*winnerR = c;
	break;
    }
    delete [] counterweights;
    delete [] active;
    delete [] talley;
}

