package org.bolson.vote;
/**
 * Interface for any voting system that inputs a voter's rankings of the choices.
 * Possible implementations: Borda count, Condorcet, IRV (sucks).
 */

public abstract class RankedVotingSystem extends VotingSystem {
    /* In addition to interfaced methods, there shall be a constructor taking 
     * the integer number N of choices to be voted on. */
    public RankedVotingSystem( int numCandidates ) {
    	super( numCandidates );
    };
    
    /**
     * Converts rating to ranking and passes to voteRanking.
     * @param ranking An array int[numc].
     *	ranking[i] is the ranking of choice i.
     */
    public int voteRating( int rating[] )  {
	int ci[] = new int[numc], rank[] = new int[numc];
	int i;
	boolean done = false;
	for ( i = 0; i < numc; i++ ) {
	    ci[i] = i;
	}
	while ( ! done ) {
	    done = true;
	    for ( i = 0; i < numc - 1; i++ ) {
		if ( rating[ci[i]] < rating[ci[i+1]] ) {
		    int t;
		    t = ci[i];
		    ci[i] = ci[i+1];
		    ci[i+1] = t;
		    done = false;
		}
	    }
	}
	for ( i = 0; i < numc; i++ ) {
	    rank[ci[i]] = i + 1;
	}
	return voteRanking( rank );
    }
    public int voteRating( float rating[] )  {
	int ci[] = new int[numc], rank[] = new int[numc];
	int i;
	boolean done = false;
	for ( i = 0; i < numc; i++ ) {
	    ci[i] = i;
	}
	while ( ! done ) {
	    done = true;
	    for ( i = 0; i < numc - 1; i++ ) {
		if ( rating[ci[i]] < rating[ci[i+1]] ) {
		    int t;
		    t = ci[i];
		    ci[i] = ci[i+1];
		    ci[i+1] = t;
		    done = false;
		}
	    }
	}
	for ( i = 0; i < numc; i++ ) {
	    rank[ci[i]] = i + 1;
	}
	return voteRanking( rank );
    }
    public int voteRating( double rating[] )  {
	int ci[] = new int[numc], rank[] = new int[numc];
	int i;
	boolean done = false;
	for ( i = 0; i < numc; i++ ) {
	    ci[i] = i;
	}
	while ( ! done ) {
	    done = true;
	    for ( i = 0; i < numc - 1; i++ ) {
		if ( rating[ci[i]] < rating[ci[i+1]] ) {
		    int t;
		    t = ci[i];
		    ci[i] = ci[i+1];
		    ci[i+1] = t;
		    done = false;
		}
	    }
	}
	for ( i = 0; i < numc; i++ ) {
	    rank[ci[i]] = i + 1;
	}
	return voteRanking( rank );
    }
    
};
