package org.bolson.vote;
import java.io.*;

/**
 * Command line app for processing votes.
 */
public class vote {
    boolean rating = true;
    boolean isFloat = false;
    boolean isBin = false;
    boolean scriptMode = false;
    boolean htmlOut = false;
    boolean printState = false;
    boolean quiet = false;
    boolean printVotes = false;
    boolean validateVotes = false;
    boolean historgam = false;
    Class fac = null;
    int numc = -1;
    boolean histogram = false;
    int histBuckets = -1;
    String names[] = null;
    int histyle = 2;
    
public void voteFromTextFile( String filename ) throws Exception {
    VotingSystem tab;
    int pi[] = null;
    float pf[] = null;
    StreamTokenizer st = null;
    DataInputStream dis = null;
    boolean goodP = true;
    int votes = 0;
    Histogram hist = null;

    if ( validateVotes && isFloat && ! rating ) {
	System.err.println("floating point rankings are nonsensical. (fractional placement?) aborting. not processing file \"" + filename + '"' );
	return;
    }
//    System.out.println("reading votes from \"" + filename + "\", processing with " + fac + ", " + numc + " choices, " + (rating ? "Rating " : "Ranking " ) + (isFloat ? "float " : "int ") + (isBin ? "bin" : "text") );
    if ( isBin ) {
	dis = new DataInputStream( new FileInputStream( filename ));
	if ( numc == -1 ) numc = dis.readInt();
    } else {
	st = new StreamTokenizer( new FileReader( filename ));
	st.eolIsSignificant( true );
	if ( numc == -1 ) {
	    if ( st.nextToken() != st.TT_NUMBER ) {
		System.out.println( filename + ": numc undefined and first token of file not a number");
		return;
	    }
	    numc = (int)st.nval;
	    if ( st.nextToken() != st.TT_EOL ) {
		System.out.println( filename + ": numc undefined and first line not a number by itself");
	    }
	}
    }
    tab = (VotingSystem)(fac.getConstructors())[0].newInstance( new Object[]{new Integer( numc )} );
    if ( isFloat ) {
	pf = new float[numc];
    } else {
	pi = new int[numc];
    }
    if ( histogram ) {
	int nbuck = histBuckets;
	double min, max;
	if ( nbuck == -1 ) {
	    if ( rating ) {
		nbuck = 11;
	    } else {
		nbuck = numc;
	    }
	}
	if ( rating ) {
	    if ( isFloat ) {
		min = -1.0;
		max = 1.0;
	    } else {
		min = 0;
		max = 10;
	    }
	} else {
	    min = 1;
	    max = numc;
	}
	hist = new Histogram( numc, nbuck, min, max );
    }
    try { while ( goodP ) {
//	System.out.print("vote:");
	for ( int i = 0; i < numc; i++ ) {
	    int err;
	    if ( isBin ) {
		if ( isFloat ) {
		    pf[i] = dis.readFloat();
		} else {
		    pi[i] = dis.readInt();
		}
	    } else {
		int tt;
		tt = st.nextToken();
		if ( tt == st.TT_EOL ) {
		    if ( i != 0 ) {
			System.err.println(filename + ": wrong number of votes on a line. eol at i="+i );
			goodP = false;
			break;
		    } else {
			tt = st.nextToken();
		    }
		}
		if ( tt != st.TT_NUMBER ) {
		    goodP = false;
		    break;
		}
		if ( isFloat ) {
		    pf[i] = (float)st.nval;
		} else {
		    pi[i] = (int)st.nval;
		}
	    }
	    if ( printVotes ) {
		if ( i != 0 ) {
		    System.out.print( '\t' );
		}
		if ( isFloat ) {
		    System.out.print( pf[i] );
		} else {
		    System.out.print( pi[i] );
		}
		System.out.print( "(" + st.nval + ")" );
	    }
	}
	if ( printVotes ) {
	    System.out.println();
	}
	if ( goodP ) {
	    if ( isFloat ) {
		if ( rating ) {
		    tab.voteRating( pf );
		    if ( hist != null ) {
			hist.addRating( pf );
		    }
		} else {
		    pi = new int[numc];
		    for ( int i = 0; i < numc; i++ ) {
			pi[i] = (int)pf[i];
		    }
		    tab.voteRanking( pi );
		    if ( hist != null ) {
			hist.addRanking( pi );
		    }
		}
	    } else {
		if ( rating ) {
		    tab.voteRating( pi );
		    if ( hist != null ) {
			hist.addRating( pi );
		    }
		} else {
		    tab.voteRanking( pi );
		    if ( hist != null ) {
			hist.addRanking( pi );
		    }
		}
	    }
	    votes++;
	}
    }} catch ( Exception e ) {
	e.printStackTrace();
    };
    //System.out.println( votes + " votes counted" );
    int[] winners = tab.getWinners();
    if ( printState ) {
	if ( names != null ) {
	    System.out.print( tab.toString( names ) );
	} else {
	    System.out.print( tab.toString() );
	}
    }
    if ( htmlOut ) {
	System.out.print( tab.htmlSummary( names ) );
    }
    if ( scriptMode ) {
	for ( int i = 0; i < winners.length; i++ ) {
	    System.out.println( winners[i] );
	}
    } else if ( winners.length == 0 ) {
	System.out.println("no winner found,"+tab);
    } else if ( winners.length == 1 ) {
	if ( names != null ) {
	    System.out.println( "Winner: " + names[winners[0]] );
	} else {
	    System.out.println("winner, choice number " + (winners[0] + 1) + " (of 1.." + numc + ")" );
	}
    } else {
	if ( names != null ) {
	    System.out.print( "Winners: " + names[winners[0]] );
	    for ( int i = 1; i < winners.length; i++ ) {
		System.out.print(", " + names[winners[i]] );
	    }
	} else {
	    System.out.print( winners.length + " winners, numbers " + (winners[0] + 1) );
	    for ( int i = 1; i < winners.length; i++ ) {
		System.out.print(", " + (winners[i] + 1) );
	    }
	}
	System.out.println(" (of 1.." + numc + ")" );
    }
    if ( hist != null ) {
	System.out.print( hist.toString( histyle, names ) );
    }
}

static final String optstring = "bBcdDfhi:n:Rrw";
static final String usageText = "vote [-bBcdfhRrw][-n number of choices][-i vote file][vote file ...]\n";
static final String helpText =
"\t-b\tbinary\n\t-B\ttext (default)\n\t-c\tCondorcet\n\t-D\tBorda count (default)\n\t-d\tinteger (default)\n\t-f\tfloat (-1.0 .. 1.0)\n\t-h\tthis help\n\t-H\toutput html table\n\t-i f\timmediately load and run vote file f\n\t-I\tInstant Runoff Vote\n\t-n N\tnumber of choices\n\t-N f\tfile with names of choices\n\t-p\tprint full state, not just winner\n\t-q\tquiet, don't print winner\n\t-R\tnumbers are rankings (1 best) (default)\n\t-r\tnumbers are ratings (higher better)\n\t-s\tscript friendly output (zero indexed \\n separated list of winners)\n\t-S\tmake histogram of ranked or rated votes\n\t-T #\thistogram style #\n\t-w\tRaw Sum of ratings\n\t-Y\tverify that votes are valid\n";

public static void main( String argv[] ) {
    vote it = new vote();
    argv = it.readArgv( argv );
    if ( argv != null ) for ( int i = 0; i < argv.length; i++ ) {
	try {
	    it.voteFromTextFile( argv[i] );
	} catch ( Exception e ) {
	    e.printStackTrace();
	}
    }
}
/**
 * @param argv like main()
 * @return unused arguments (filenames)
 */
public String[] readArgv( String argv[] ) {
    int c;
    int i;

    try {
	fac = Class.forName("org.bolson.vote.RawRating");
    } catch ( java.lang.ClassNotFoundException e ) {
	e.printStackTrace();
    }
    for ( i = 0; i < argv.length; i++ ) {
	if ( argv[i].equals("-B") ) {
	    isBin = false;
	} else if ( argv[i].equals("-b") ) {
	    isBin = true;
	} else if ( argv[i].equals("-c") ) {
	    try {
		fac = Class.forName("org.bolson.vote.Condorcet");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-D") ) {
	    try {
		fac = Class.forName("org.bolson.vote.BordaVotingSystem");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-d") ) {
	    isFloat = false;
	} else if ( argv[i].equals("-f") ) {
	    isFloat = true;
	} else if ( argv[i].equals("-h") ) {	// help
	    System.out.println(usageText);
	    System.out.println(helpText);
	} else if ( argv[i].equals("-H") ) {
	    htmlOut = true;
	} else if ( argv[i].equals("-I") ) {
	    try {
		fac = Class.forName("org.bolson.vote.IRV");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-i") ) {
	    i++;
	    try {
		voteFromTextFile( argv[i] );
	    } catch ( Exception e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-N") ) {
	    i++;
	    names = readNames( argv[i] );
	} else if ( argv[i].equals("-n") ) {
	    i++;
	    numc = Integer.parseInt( argv[i] );
	} else if ( argv[i].equals("-p") ) {
	    printState = true;
	} else if ( argv[i].equals("-q") ) {
	    quiet = true;
	} else if ( argv[i].equals("-R") ) {
	    rating = false;
	} else if ( argv[i].equals("-r") ) {
	    rating = true;
	} else if ( argv[i].equals("-s") ) {
	    scriptMode = true;
	} else if ( argv[i].equals("-S") ) {
	    histogram = true;
	} else if ( argv[i].equals("-T") ) {
	    i++;
	    histyle = Integer.parseInt( argv[i] );
	    histogram = true;
	} else if ( argv[i].equals("-w") ) {
	    try {
		fac = Class.forName("org.bolson.vote.RawRating");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-Y") ) {
	    validateVotes = true;
	} else if ( argv[i].equals("--system") ) {
	    i++;
	    try {
		fac = Class.forName(argv[i]);
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("--printVotes") ) {
	    printVotes = true;
	} else if ( argv[i].equals("--") ) {
	    break;
	} else if ( argv[i].startsWith("-") ) {
	    System.err.println("bogus flag: " + argv[i] );
	    return null;
	} else {
	    break;
	}
    }
    String toret[] = new String[argv.length - i];
    for ( int j = 0; i + j < argv.length; j++ ) {
	toret[j] = argv[i+j];
    }
    return toret;
}

/**
 * @param filename A file with the names of choices, one per line.
 * @return The names (lines) of the file.
 */
static String[] readNames( String filename ) {
    try {
	FileReader fr = new FileReader( filename );
	BufferedReader br = new BufferedReader( fr );
	java.util.Vector tn = new java.util.Vector();
	String s;
	while ( (s = br.readLine()) != null ) {
	    tn.addElement( s );
	}
	br.close();
	String[] toret = new String[tn.size()];
	for ( int i = 0; i < toret.length; i++ ) {
	    toret[i] = (String)tn.elementAt( i );
	}
	return toret;
    } catch ( Exception e ) {
	return null;
    }
}
}
