/*
 *  Choice.cpp
 *  multiseat
 *
 *  Created by Brian Olson on Sun Oct 12 2003.
 *  Copyright (c) 2003 Brian Olson. All rights reserved.
 *
 */

#include "Choice.h"

#include <stdlib.h>
#include <stdio.h>

void Choice::runElection( int* winnerR, int seats, const VoterArray& they ) {
    bool* active;
    double* strength;
    int* curvote;
    double* talley;
    int winCount = 0;
    double droop;

    active = new bool[they.numc];
    strength = new double[they.numv];
    curvote = new int[they.numv];
    talley = new double[they.numc];

    droop = 1.0 + ( ((double)they.numv) / ( ((double)seats) + 1.0 ) );

    for ( int v = 0; v < they.numv; v++ ) {
	strength[v] = 1.0;
    }
    for ( int c = 0; c < they.numc; c++ ) {
	active[c] = true;
    }

    for ( int y = 0; y < they.numc; y++ ) {
	/* count first place votes
	 * fractional if vote partially used already */
	for ( int v = 0; v < they.numv; v++ ) if ( strength[v] > 0.0 ) {
	    int fav = -1;
	    double fperf = -2.0;
	    Voter &cv = they[v];
	    for ( int c = 0; c < they.numc; c++ ) if ( active[c] ) {
		double p = cv.getPref(c);
		if ( p > fperf ) {
		    fperf = p;
		    fav = c;
		}
	    }
	    if ( fav != -1 ) {
		curvote[v] = fav;
		talley[fav] += strength[v];
	    }
	}
	/* see who won OR see who lost, deactivate that candidate */
	{
	double tmin = they.numv, tmax = 0.0;
	int mini = -1, maxi = -1;
	/* find highest and lowest */
	for ( int c = 0; c < they.numc; c++ ) if ( active[c] ) {
	    if ( talley[c] < tmin ) {
		tmin = talley[c];
		mini = c;
	    }
	    if ( talley[c] > tmax ) {
		tmax = talley[c];
		maxi = c;
	    }
	}
	if ( tmax >= droop ) {
	    active[maxi] = false;
	    winnerR[winCount] = maxi;
	    winCount++;
	} else if ( mini != -1 ) {
	    active[mini] = false;
	} else {
	    fprintf(stderr,"found neither min nor max\n");
	    abort();
	}
	}
    }
    delete active;
    delete strength;
}

MultiseatElection* newChoice() {
    return new Choice();
}


static MEFactory ChoiceFactory( newChoice, "choice" );

const char* Choice::name() {
    return "Choice";
}
