/*
 *  MaxHC.cpp
 *  multiseat
 *
 *  Created by Brian Olson on Mon Oct 13 2003.
 *  Copyright (c) 2003 Brian Olson. All rights reserved.
 *
 */

#include "MaxHC.h"

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

void MaxHC::runElection( int* winnerR, int seats, const VoterArray& they ) {
    int numc = they.numc;
    int* bestWinnerR;
    double bestHC = -1.0;
    double stddevP[2];
    bool notdone = true;

    bestWinnerR = new int[numc];

    for ( int c = 0; c < seats; c++ ) {
	winnerR[c] = bestWinnerR[c] = c;
    }
    for ( int c = seats; c < numc; c++ ) {
	winnerR[c] = bestWinnerR[c] = -1;
    }
    while ( notdone ) {
	double th;
	/* try this configuration of winnerR */
	th = MultiseatElection::happiness( they, they.numv, winnerR, seats, stddevP );
#if 01
	th = pow( th, hpow );
	th /= pow( stddevP[0], cpow );
#else
	th /= stddevP[0];
#endif
	if ( th > bestHC ) {
	    bestHC = th;
	    for ( int s = 0; s < seats; s++ ) {
		bestWinnerR[s] = winnerR[s];
	    }
	}
	{
	    /* chose next winnerR configuration */
	    int incpos = seats - 1;
	    winnerR[incpos]++;
	    while ( winnerR[incpos] >= numc - (seats - 1 - incpos) ) {
		winnerR[incpos] = 0;
		if ( incpos == 0 ) {
		    notdone = false;
		    break;
		}
		incpos--;
		winnerR[incpos]++;
	    }
	    while ( incpos < seats - 1 ) {
		winnerR[incpos+1] = winnerR[incpos] + 1;
		incpos++;
	    }
#if 0
	    for ( int c = 0; c < seats; c++ ) {
		printf("%d ", winnerR[c] );
		assert( winnerR[c] >= 0 && winnerR[c] < numc );
	    }printf("\n");
#endif
	}
    }
    for ( int s = 0; s < seats; s++ ) {
	winnerR[s] = bestWinnerR[s];
    }
    delete bestWinnerR;
}

MultiseatElection* newMaxHC() {
    return new MaxHC();
}

static MEFactory randomFactory( newMaxHC, "maxhc" );

const char* MaxHC::name() {
    static char hcname[64];
    if ( hpow == 1.0 ) {
	if ( cpow == 1.0 ) {
	    return "Maximized Happiness*Consensus";
	}
	sprintf( hcname, "Max Happiness*(Consensus^%f)", cpow );
    } else {
	if ( cpow == 1.0 ) {
	    sprintf( hcname, "Max (Happiness^%f)*Consensus", hpow );
	} else {
	    sprintf( hcname, "Max (Happiness^%f)*(Consensus^%f)", hpow, cpow );
	}
    }
    return hcname;
}

MaxHC::MaxHC()
: hpow( 1.0 ), cpow( 1.0 )
{}

void MaxHC::init( const char** envp ) {
    if ( envp == NULL )
	return;
    for ( int i = 0; envp[i]; i++ ) {
	if ( ! strcmp( envp[i], "-h" ) ) {
	    i++;
	    hpow = atof( envp[i] );
	} else if ( ! strcmp( envp[i], "-c" ) ) {
	    i++;
	    cpow = atof( envp[i] );
	} else {
	    fprintf(stderr,"MaxHC ignores \"%s\"\n", envp[i] );
	}
    }
}
