/*
 *  MultiseatElection.cpp
 *  multiseat
 *
 *  Created by Brian Olson on Mon Sep 08 2003.
 *  Copyright (c) 2003 Brian Olson. All rights reserved.
 *
 */

#include "MultiseatElection.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>

void MultiseatElection::init( const char** envp ) {
}

/* calculate average happiness of the voters for the winner */
double MultiseatElection::happiness( const VoterArray& they, int numv, int* winner, int seats, double* stddevP, int start ) {
    double candAvg = 0, voterAvg, avgVoterStd = 0;
    int end = start + numv;
    for ( int v = start; v < end; v++ ) {
	Voter &cv = they[v];
	double vstd = 0;
	voterAvg = 0;
	for ( int s = 0; s < seats; s++ ) {
	    double td = cv.getPref( winner[s] );
	    candAvg += td;
	    voterAvg += td;
	}
	voterAvg = voterAvg / seats;
	for ( int s = 0; s < seats; s++ ) {
	    double td = cv.getPref( winner[s] );
	    td = td - voterAvg;
	    vstd += td * td;
	}
	vstd = vstd / seats;
	avgVoterStd += sqrt( vstd );
    }
    candAvg = candAvg / numv;
    if ( stddevP != NULL ) {
	double stddev = 0;
	avgVoterStd = avgVoterStd / numv;
	stddevP[1] = avgVoterStd;
	for ( int v = start; v < end; v++ ) {
	    Voter &cv = they[v];
	    voterAvg = 0;
	    for ( int s = 0; s < seats; s++ ) {
		double td = cv.getPref( winner[s] );
		voterAvg += td;
	    }
	    voterAvg = voterAvg - candAvg;
	    stddev += ( voterAvg * voterAvg );
	}
	stddev = stddev / numv;
	stddev = sqrt( stddev );
	stddevP[0] = stddev;
    }
    return candAvg;
}

#if 0
double MultiseatElection::happiness( const VoterArray& they, int numv, int* winner, int seats, double* stddevP ) {
    return happiness( they, numv, winner, seats, stddevP, 0 );
}
#endif

double MultiseatElection::happiness( const VoterArray& they, int numv, int* winner, int seats ) {
    return happiness( they, numv, winner, seats, NULL );
}

double MultiseatElection::happiness( const VoterArray& they, int* winner, int seats ) {
    return happiness( they, they.numv, winner, seats );
}


MEFactory* MEFactory::root = NULL;

MEFactory::MEFactory( MultiseatElection* (*f)(), char* n )
: fact( f ), name( n )
{
    next = root;
    root = this;
}

const MEFactory* MEFactory::byName( const char* n ) {
    MEFactory* cur = root;
    while ( cur != NULL ) {
	if ( ! strcmp( n, cur->name ) ) {
	    return cur;
	}
	cur = cur->next;
    }
    return NULL;
}

MESystem* MESystem::newMESystem( const MEFactory* f, const char** a, MESystem* n ) {
    if ( f == NULL ) {
	return n;
    }
    MESystem* toret = new MESystem( f, a, n );
    return toret;
}

MESystem* MESystem::defaultList( const char** a, MESystem* n ) {
    MEFactory* cur = MEFactory::root;
    MESystem* toret = n;
    while ( cur != NULL ) {
	toret = new MESystem( cur, a, toret );
	cur = cur->next;
    }
    return toret;
}

MESystem::~MESystem() {
    if ( next ) delete next;
    if ( args ) delete args;
}

void MESystem::init() {
    me = fact->make();
    if ( args ) {
	me->init( args );
    }
}

void MESystem::print( void* vf ) {
    FILE* f = (FILE*)vf;
    fprintf( f, "fatory \"%s\", ", fact->name );
    if ( args != NULL ) {
	fprintf( f, "args \"%s", args[0] );
	for ( int i = 1; args[i] != NULL; i++ ) {
	    fprintf( f, "\", \"%s", args[i] );
	}
	fprintf( f, "\"" );
    } else {
	fprintf( f, "no args" );
    }
    fprintf( f, "\n" );
}

