#ifndef MULTISEAT_ELECTION_H
#define MULTISEAT_ELECTION_H
/*
 *  MultiseatElection.h
 *  multiseat
 *
 *  Created by Brian Olson on Mon Sep 08 2003.
 *  Copyright (c) 2003 Brian Olson. All rights reserved.
 *
 */

#include "Voter.h"

class MultiseatElection {
public:
    /*!
     * returns -1 terminated winner list in winnerR.
     * winnerR should be int[seats].
     */
    virtual void runElection( int* winnerR, int seats, const VoterArray& they ) = 0;

    /*!
     * null terminated array of char*, default action is to ignore all.
     * can take arbitrary options to vary election method.
     */
    virtual void init( const char** envp );

    /*!
     * return a name for this election, possibly including effects of envp
     */
    virtual const char* name() = 0;

    /* calculate average happiness of the voters for the winner */
    static double happiness( const VoterArray& they, int numv, int* winner, int seats, double* stddevP, int start=0 );
//    static double happiness( const VoterArray& they, int numv, int* winner, int seats, double* stddevP );
    static double happiness( const VoterArray& they, int numv, int* winner, int seats );
    static double happiness( const VoterArray& they, int* winner, int seats );
};

class MEFactory {
public:
    MultiseatElection* (*fact)();
    char* name;

    inline MultiseatElection* make() const {
	return fact();
    }

    MEFactory* next;
    static MEFactory* root;

    //MEFactory();
    MEFactory( MultiseatElection* (*f)(), char* n );

    static const MEFactory* byName( const char* n );
};

class MESystem {
public:
    const MEFactory* fact;
    const char** args;
    MESystem* next;

    MultiseatElection* me;

    inline MESystem( const MEFactory* f, const char** a, MESystem* n = (MESystem*)0 )
	: fact( f ), args( a ), next( n )
    {}
    /*!
	* a list builting factory. root = newMESystem( f, a, n );
     */
    static MESystem* newMESystem( const MEFactory* f, const char** a, MESystem* n = (MESystem*)0 );
    /*!
	* a list builting factory. root = newMESystem( f, a, n );
     */
    inline static MESystem* newMESystem( const char* factoryName, const char** a, MESystem* n = (MESystem*)0 ) {
	return newMESystem( MEFactory::byName( factoryName ), a, n );
    }

    /*!
     * return one of each in the factor list
     */
    static MESystem* defaultList( const char** a, MESystem* n );
    inline static MESystem* defaultList() {
	return MESystem::defaultList( (const char**)0, (MESystem*)0 );
    }

    ~MESystem();

    void print( void* f);

    void init();

    inline void runElection( int* winnerR, int seats, const VoterArray& they ) {
	me->runElection( winnerR, seats, they );
    }

    inline const char* name() {
	return me->name();
    }
};

#endif
