/*
 *  NormRatedCIR.cpp
 *  multiseat
 *
 *  Created by Brian Olson on Sun Oct 12 2003.
 *  Copyright (c) 2003 Brian Olson. All rights reserved.
 *
 */

#include "NormRatedCIR.h"

#include <math.h>
#include <stdlib.h>
#include <stdio.h>

void NormRatedCIR::runElection( int* winnerR, int seats, const VoterArray& they ) {
    bool* active;
    double* strength;
    int* curvote;
    double* psum;
    double droop;

    active = new bool[they.numc];
    strength = new double[they.numv];
    curvote = new int[they.numv];
    psum = new double[they.numc];

    droop = 1.0 + ( ((double)they.numv) / ( ((double)seats) + 1.0 ) );

    for ( int v = 0; v < they.numv; v++ ) {
	strength[v] = 1.0;
    }
    for ( int c = 0; c < they.numc; c++ ) {
	active[c] = true;
    }

    for ( int stillEnabled = they.numc; stillEnabled != seats; stillEnabled--) {
	/* add up normalized votes
	 * fractional if vote partially used already */
	for ( int v = 0; v < they.numv; v++ ) if ( strength[v] > 0.0 ) {
	    double abssum = 0.0;
	    Voter &cv = they[v];
	    for ( int c = 0; c < they.numc; c++ ) if ( active[c] ) {
		double p = cv.getPref(c);
		abssum += fabs( p );
	    }
	    for ( int c = 0; c < they.numc; c++ ) if ( active[c] ) {
		double p = cv.getPref(c);
		psum[c] += p / abssum;
	    }
	}
	/* disqualify the loser */
	{
	    double tmin = they.numv + 1.0;
	    int mini = -1;
	    /* find lowest */
	    for ( int c = 0; c < they.numc; c++ ) if ( active[c] ) {
		if ( psum[c] < tmin ) {
		    tmin = psum[c];
		    mini = c;
		}
	    }
	    if ( mini == -1 ) {
		fprintf(stderr,"found no min\n");
		abort();
	    }
	    active[mini] = false;
	}
    }
    {
	/* write out results */
	int wRpos = 0;
	for ( int c = 0; c < they.numc; c++ ) if ( active[c] ) {
	    winnerR[wRpos] = c;
	    wRpos++;
	}
	for ( int c = 0; c < they.numc; c++ ) if ( ! active[c] ) {
	    winnerR[wRpos] = c;
	    wRpos++;
	}
    }
    delete active;
    delete strength;
}

MultiseatElection* newNormRatedCIR() {
    return new NormRatedCIR();
}


static MEFactory NormRatedCIRFactory( newNormRatedCIR, "nrcir" );

const char* NormRatedCIR::name() {
    return "Normalized Rating, Instant Runoff";
}
