/*
 *  PickN.cpp
 *  multiseat
 *
 *  Created by Brian Olson on Sun Oct 12 2003.
 *  Copyright (c) 2003 Brian Olson. All rights reserved.
 *
 */

#include "PickN.h"

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

void PickN::runElection( int* winnerR, int seats, const VoterArray& they ) {
    double* psum;
    int* rank;
    bool notdone = true;
    int numc = they.numc;
    unsigned long* talley;
    int pn = (pickN == -1) ? seats : pickN;

    psum = new double[numc];
    rank = new int[numc];
    talley = new unsigned long[numc];
    for ( int c = 0; c < they.numc; c++ ) {
	talley[c] = 0;
    }

    // add up votes
    for ( int v = 0; v < they.numv; v++ ) {
	for ( int s = 0; s < pn; s++ ) {
	    psum[s] = -2.0;
	    rank[s] = -1;
	}
	// mini sort to find top N
	Voter &cv = they[v];
	for ( int c = 0; c < they.numc; c++ ) {
	    double p = cv.getPref(c);
	    int sb;
	    sb = 0;
	    while ( p > psum[sb] && sb < pn ) {
		if ( sb != 0 ) {
		    psum[sb-1] = psum[sb];
		    rank[sb-1] = rank[sb];
		}
		psum[sb] = p;
		rank[sb] = c;
		sb++;
	    }
	}
	// vote for top N
	for ( int s = 0; s < pn; s++ ) {
	    assert( rank[s] != -1 );
	    talley[rank[s]]++;
	}
    }
    for ( int s = 0; s < numc; s++ ) {
	rank[s] = s;
    }
    /* sort candidates based on vote talley */
    while ( notdone ) {
	notdone = false;
	for ( int s = 1; s < numc; s++ ) {
	    if ( talley[rank[s-1]] < talley[rank[s]] ) {
		int tr;
		tr = rank[s-1];
		rank[s-1] = rank[s];
		rank[s] = tr;
		notdone = true;
	    }
	}
    }
    for ( int s = 0; s < numc; s++ ) {
	winnerR[s] = rank[s];
    }
    delete talley;
    delete psum;
    delete rank;
}

void PickN::init( const char** envp ) {
    if ( envp == NULL )
	return;
    for ( int i = 0; envp[i]; i++ ) {
	if ( ! strcmp( envp[i], "-n" ) ) {
	    i++;
	    pickN = atoi( envp[i] );
	} else {
	    fprintf(stderr,"PickN ignores \"%s\"\n", envp[i] );
	}
    }
}

PickN::PickN()
: pickN( -1 ) {
}

MultiseatElection* newPickN() {
    return new PickN();
}


static MEFactory PickNFactory( newPickN, "pickn" );

const char* PickN::name() {
    if ( pickN == -1 ) {
	return "Votes = Seats";
    } else {
	static char pnName[] = "Pick NNNNNN";
	snprintf( pnName + 5, 6, "%d", pickN );
	return pnName;
    }
}

