/*
 *  SimWorkState.cpp
 *  multiseat
 *
 *  Created by Brian Olson on Thu Sep 11 2003.
 *  Copyright (c) 2003 Brian Olson. All rights reserved.
 *
 */

#include "MultiseatElection.h"
#include "SimWorkState.h"
#include "SimGlobal.h"
#include "DBResultFile.h"
#include "WorkQueue.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifndef NULL
#define NULL (0)
#endif

SimWorkState::SimWorkState()
: g( NULL ), q( NULL ), happiness( NULL ), happisum( NULL ), happistdsum( NULL ),
  winnerR( NULL )
{}

void SimWorkState::init() {
    int nsys = g->nsys;
    int sys;

    trials = 0;
    happisum = new double[nsys];
    happistdsum = new double[nsys];
    happiness = new double*[nsys];
    for ( sys = 0; sys < nsys; sys++ ) {
        happiness[sys] = NULL;
    }
#if 0
    /* depends on numv, move to per-wu function */
    if ( g->strategies ) {
	int curcount = 0;
	for ( int i = 0; i < g->numStrat; i++ ) {
	    if ( strategies[i]->count == -1 ) {
		strategies[i]->count = numv / numStrat;
	    }
	    curcount += strategies[i]->count;
	    strategies[i]->happisum = new double[nsys];
	    strategies[i]->happistdsum = new double[nsys];
	    strategies[i]->happiness = new double*[nsys];
	    for ( sys = 0; sys < nsys; sys++ ) {
		strategies[i]->happiness[sys] = new double[trials];
	    }
	}
	strategies[numStrat-1]->count += numv - curcount;
    }
#endif
}

void SimWorkState::runWu( WorkUnit* wu ) {
    Result* r = NULL;
    Result* nr = NULL;
    int sys;
    int nsys = g->nsys;
    VoterArray theyWithError;
    double stddevP[2];

    nr = newResult( nsys * (g->numStrat + 1) );
    if ( g->drf != NULL ) {
	r = g->drf->get( wu->numc, wu->numv, wu->seats, wu->error );
	if ( r ) {
	    memcpy( nr, r, sizeofRusult( nsys * (g->numStrat + 1) ) );
	}
    }
    if ( g->upToTrials ) {
	if ( g->upToTrials <= nr->trials ) {
	    free( nr );
	    return;
	}
	if ( g->upToTrials < wu->trials + nr->trials ) {
	    wu->trials = g->upToTrials - nr->trials;
	}
    }
    if ( wu->trials > trials ) {
	for ( sys = 0; sys < nsys; sys++ ) {
	    if ( happiness[sys] ) {
		delete happiness[sys];
	    }
	    happiness[sys] = new double[wu->trials];
	}
	trials = wu->trials;
    }
    if ( winnerR ) {
	free( winnerR );
    }
    winnerR = (int*)malloc(sizeof(int)*wu->numc);
    they.build( wu->numv, wu->numc );
    // main work here
    for ( int t = 0; t < trials; t++ ) {
	if ( g->goGently ) return;
	for ( int v = 0; v < wu->numv; v++ ) {
	    they[v].randomize();
	}
	int sys;
	MESystem* cur;
	for ( cur = g->root, sys = 0; cur; cur = cur->next, sys++ ) {
	    if ( g->goGently ) return;
	    cur->runElection( winnerR, wu->seats, they );
	    happisum[sys] += happiness[sys][t] = MultiseatElection::happiness( they, wu->numv, winnerR, wu->seats, stddevP );
	    happistdsum[sys] += stddevP[0];
	    if ( g->printEachElection ) { printElection( cur, wu, happiness[sys][t] ); }
	}
    }
    // calculate run stats
    unsigned int ttot = trials + nr->trials;
    for ( sys = 0; sys < nsys; sys++ ) {
	double variance;
	nr->systems[sys].meanHappiness = (happisum[sys] + nr->systems[sys].meanHappiness * nr->trials) / ttot;
	nr->systems[sys].consensusStdAvg = (happistdsum[sys] + nr->systems[sys].consensusStdAvg * nr->trials) / ttot;
	happistdsum[sys] /= trials;
	variance = nr->systems[sys].reliabilityStd * nr->systems[sys].reliabilityStd * nr->trials;
	for ( int i = 0; i < trials; i++ ) {
	    double d;
	    d = happiness[sys][i] - nr->systems[sys].meanHappiness;
	    variance += d * d;
	}
	variance /= ttot;
	nr->systems[sys].reliabilityStd = sqrt( variance );
    }
#if 0
    if ( strategies ) for ( int st = 0; st < numStrat; st++ ) {
	int rsys = nsys * (st + 1);
	for ( sys = 0; sys < nsys; sys++ ) {
	    double variance;
	    nr->systems[rsys+sys].meanHappiness = (strategies[st]->happisum[sys] + nr->systems[rsys+sys].meanHappiness * nr->trials) / ttot;
	    nr->systems[rsys+sys].consensusStdAvg = (strategies[st]->happistdsum[sys] + nr->systems[rsys+sys].consensusStdAvg * nr->trials) / ttot;
	    strategies[st]->happistdsum[sys] /= trials;
	    variance = nr->systems[rsys+sys].reliabilityStd * nr->systems[rsys+sys].reliabilityStd * nr->trials;
	    for ( i = 0; i < trials; i++ ) {
		double d;
		d = strategies[st]->happiness[sys][i] - nr->systems[rsys+sys].meanHappiness;
		variance += d * d;
	    }
	    variance /= ttot;
	    nr->systems[rsys+sys].reliabilityStd = sqrt( variance );
	}
    }
#endif
    nr->trials = ttot;
    // finish up
    if ( g->summary ) {
	int sys;
	MESystem* cur;
	wu->printSummaryHeader( stdout, nr->trials, g->summary );
//	printResult( stdout, nr, /*g->root*/NULL, nsys, g->summary );
	for ( cur = g->root, sys = 0; cur; cur = cur->next, sys++ ) {
	    printSystemResult( stdout, nr->systems + sys, cur->name(), g->summary );
	}
#if 0
	if ( g->strategies ) for ( int st = 0; st < g->numStrat; st++ ) {
	    if ( summary == 4 ) {
		fprintf( stdout, "<tr><td colspan=\"3\" bgcolor=\"#cccccc\">Strategy \"%s\":</td></tr>\n", strategies[st]->name() );
	    } else if ( summary != 0 ) {
		fprintf( stdout, "strategy \"%s\":\n", strategies[st]->name() );
	    }
	    for ( int i = 0; i < nsys; i++ ) {
		printSystemResult( stdout, nr->systems + ((nsys*(st+1)) + i), nb.names ? nb.names[i] : NULL, summary );
	    }
	}
#endif
	wu->printSummaryFooter( stdout, nr->trials, g->summary );
    }
    if ( g->drf ) {
	g->drf->put( nr, wu->numc, wu->numv, wu->seats, wu->error );
    }
    free( nr );
}

void* SimWorkState::run() {
    WorkUnit* wu = (WorkUnit*)0;

    while ( (!g->goGently) && (wu = q->newWorkUnit( wu )) ) {
	runWu( wu );
    }

    return (void*)0;
}

void SimWorkState::printElection( MESystem* cur, WorkUnit* wu, double happy ) {
    printf("%s\thappiness %f\twinners: %d", cur->name(), happy, winnerR[0] );
    for ( int i = 1; i < wu->seats; i++ ) {
	printf(", %d", winnerR[i] );
    }
    printf("\t");
    wu->print(stdout);
}

void* runSimWorkState( /*SimWorkState*/void* s ) {
    return ((SimWorkState*)s)->run();
}

