#ifndef WORK_QUEUE_H
#define WORK_QUEUE_H

#include <pthread.h>
#include <stdio.h>

class WorkUnit {
public:
    int numv;
    int numc;
    int seats;
    float error;
    int trials;

    WorkUnit* next;

    WorkUnit( int nv, int nc, int ns, float e, int t );

    inline void set( int nv, int nc, int ns, float e, int t ) {
	numv = nv;
	numc = nc;
	seats = ns;
	error = e;
	trials = t;
    }

    void print( FILE* file );
    void printSummaryHeader( FILE* file, int totalTrials, int style );
    void printSummaryFooter( FILE* file, int totalTrials, int style );
};

class WorkSource {
public:
    virtual WorkUnit* newWorkUnit( WorkUnit* ) = 0;
    virtual WorkUnit* newWorkUnit();
};

class WorkQueue : public WorkSource {
public:
    WorkQueue();
    virtual ~WorkQueue();
    
    void add( WorkUnit* it );
    WorkUnit* pop();
    virtual WorkUnit* newWorkUnit( WorkUnit* ); // calls pop

protected:
    WorkUnit* head;
    WorkUnit* tail;
};

class ThreadSafeWorkSource : public WorkSource {
public://protected:
    pthread_mutex_t m;
    WorkSource* realSource;
public:
    ThreadSafeWorkSource( WorkSource* source );
    virtual ~ThreadSafeWorkSource();
    virtual WorkUnit* newWorkUnit( WorkUnit* );
};

class Steps : public WorkSource {
public:
    virtual WorkUnit* newWorkUnit( WorkUnit* );
//protected:
    int* vsteps;
    int vstepsLen;
    int* csteps;
    int cstepsLen;
    int* seatsteps;
    int seatstepsLen;
    float* esteps;
    int estepsLen;
    int v, c, s, e;
public:
    Steps();
    Steps( const char* vlist, const char* clist, const char* seatlist, const char* elist, int nIn );
    virtual ~Steps();
    void parseV( const char* vlist );
    void parseC( const char* clist );
    void parseSeats( const char* seatlist );
    void parseE( const char* elist );
    void addIfNotPresent( int v, int c, int s, float e );
    int n;

    void print( void* );
    void print();
};

#endif
