#include "Random.h"
#include "Max.h"
#include "WorkQueue.h"
#include "DBResultFile.h"
#include "ThreadSafeDBRF.h"
#include "SimGlobal.h"
#include "SimWorkState.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef MAX_METHOD_ARGS
#define MAX_METHOD_ARGS 256
#endif

static SimGlobal gStorage;
static SimGlobal* g = &gStorage;
static void mysigint( int a ) {
    g->goGently = 1;
}

int main( int argc, const char* argv[] ) {
    int i;
    const char** methodArgs;
    int methodArgc;
    int printSys = 1;
    Steps* stepq = NULL;
    const char* drfname = NULL;
    int numThreads = 1;

    srandom(time(NULL));

    methodArgs = new const char*[MAX_METHOD_ARGS];
    methodArgc = 0;
    i = 0;
    while ( i < argc ) {
	const char* arg;
	arg = argv[i];
	if ( arg[0] == '-' ) {
	    if ( arg[1] == 'c' && arg[2] == '\0' ) {
		// clear out default Sys List
		if ( g->root != NULL ) {
		    delete g->root;
		}
		g->root = NULL;
	    } else if ( arg[1] == 'M' && arg[2] == '\0' ) {
		i++; arg = argv[i];
		if ( methodArgc < MAX_METHOD_ARGS ) {
		    methodArgs[methodArgc] = arg;
		    methodArgc++;
		} else {
		    fprintf( stderr, "arg \"%s\" Is beyond limit of %d method args\n", arg, MAX_METHOD_ARGS );
		}
	    } else if ( arg[1] == 'e' && arg[2] == '\0' ) {
		// enable a system, commit its args
		i++;
		arg = argv[i];
		if ( methodArgc > 0 ) {
		    const char** marg = new const char*[methodArgc+1];
		    for ( int j = 0; j < methodArgc; j++ ) {
			marg[j] = methodArgs[j];
		    }
		    marg[methodArgc] = NULL;
		    g->root = MESystem::newMESystem( arg, marg, g->root );
		} else {
		    g->root = MESystem::newMESystem( arg, NULL, g->root );
		}
		methodArgc = 0;
	    } else if ( arg[1] == 'q' && arg[2] == '\0' ) {
		printSys = 0;
	    } else if ( strcmp( arg + 1, "printSys" ) == 0 ) {
		printSys = 1;
	    } else if ( arg[1] == 'n' && arg[2] == '\0' ) {
		i++;
		if ( stepq == NULL ) stepq = new Steps;
		stepq->n = atoi( argv[i] );
	    } else if ( arg[1] == 'F' && arg[2] == '\0' ) {
		i++;
		drfname = argv[i];
	    } else if ( arg[1] == '-' ) {
		if ( ! strcmp( arg + 2, "vsteps" ) ) {
		    i++;
		    if ( stepq == NULL ) stepq = new Steps;
		    stepq->parseV( argv[i] );
		} else if ( ! strcmp( arg + 2, "csteps" ) ) {
		    i++;
		    if ( stepq == NULL ) stepq = new Steps;
		    stepq->parseC( argv[i] );
		} else if ( ! strcmp( arg + 2, "esteps" ) ) {
		    i++;
		    if ( stepq == NULL ) stepq = new Steps;
		    stepq->parseE( argv[i] );
		} else if ( (! strcmp( arg + 2, "seatsteps" )) || (! strcmp( arg + 2, "ssteps" )) ) {
		    i++;
		    if ( stepq == NULL ) stepq = new Steps;
		    stepq->parseSeats( argv[i] );
		} else if ( ! strcmp( arg + 2, "iter" ) ) {
		    i++;
		    if ( stepq == NULL ) stepq = new Steps;
		    stepq->n = atoi( argv[i] );
		} else if ( ! strcmp( arg + 2, "defSys" ) ) {
		    g->root = MESystem::defaultList( NULL, g->root );
		} else if ( ! strcmp( arg + 2, "printEachElection" ) ) {
		    g->printEachElection = 1;
		} else {
		    goto badarg;
		}
	    } else {
		badarg:
		fprintf( stderr, "bogus argv[%d] \"%s\"\n", i, arg );
		exit( 1 );
	    }
	}
	i++;
    }
    if ( g->root == NULL ) {
//	g->root = defaultSysList();
	g->root = MESystem::defaultList();
    } else {
	// reverse list which was built as a stack so that it is in user order.
	g->reverseSystems();
    }

    MESystem* cursys = g->root;
    g->nsys = 0;
    while ( cursys != NULL ) {
	g->nsys++;
	cursys->init();
	if ( printSys ) {
	    cursys->print( stdout );
	}
	cursys = cursys->next;
    }
    NameBlock nb;
    printf("%d systems\n", g->nsys );
    if ( stepq == NULL ) {
	stepq = new Steps();
    }
    stepq->print( stdout );

    if ( drfname != NULL ) {
	printf("opening result db \"%s\"...", drfname );
	if ( numThreads == 1 ) {
	    g->drf = DBResultFile::open( drfname );
	} else {
	    g->drf = ThreadSafeDBRF::open( drfname );
	}
	if ( g->drf ) {
	    printf("success\n");
	    g->drf->useNames( &nb );
	} else {
	    printf("failed\n");
	}
    }

    signal( SIGINT, mysigint );

    SimWorkState* s = new SimWorkState[numThreads];
    for ( int i = 0; i < numThreads; i++ ) {
	s[i].g = g;
	s[i].init();
    }
    
    if ( numThreads == 1 ) {
	printf("running work set\n");
	s[0].q = stepq;
	runSimWorkState( s + 0 );
    } else {
	WorkSource* q;
	pthread_t* threads;
	printf("running %d threads...", numThreads);
	q = new ThreadSafeWorkSource( stepq );
	threads = new pthread_t[numThreads];
	for ( int i = 0; i < numThreads; i++ ) {
	    s[i].q = q;
	    pthread_create( threads + i, NULL, runSimWorkState, s + i );
	}
	printf("running\n");
	for ( int i = 0; i < numThreads; i++ ) {
	    pthread_join( threads[i], NULL );
	    printf("thread %d done\n", i );
	}
	printf("all threads done\n");
	delete [] threads;
	delete q;
    }

    if ( g->drf ) {
	g->drf->flush();
	g->drf->close();
    }
    return 0;
}
