#include "FuzzyVotePickOne.h"
#include "Voter.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// inline function to avoid double evaluation of a or b
inline double fmax( double a, double b ) {
    return (a>b)?a:b;
}

double quantize( double v, int steps );

FuzzyVotePickOne::FuzzyVotePickOne( char* name, int doCounterweightI, int quantizeI )
    : VotingSystem( name ), doCounterweight( doCounterweightI ), quantization(quantizeI)
{
};

void FuzzyVotePickOne::runElection( int* winnerR, const VoterArray& they ) {
    int i;
    double* talley;
    int winner;
    double counterweight;
    int numc = they.numc;
    int numv = they.numv;

    // init things
    talley = new double[numc];
    for ( i = 0; i  < numc; i++ ) {
        talley[i] = 0;
    }

    // sum weighted votes
    switch ( doCounterweight ) {
    case 2:
	// maximize so that strongest like/dislike is 1.0/-1.0
	for ( i = 0; i < numv; i++ ) {
	    int j;
            counterweight = 0.0;
            for ( j = 0; j < numc; j++ ) {
                counterweight = fmax( counterweight, fabs( they[i].getPref(j) ));
            }
            counterweight = 1/counterweight;
            for ( j = 0; j < numc; j++ ) {
                talley[j] += they[i].getPref(j) * counterweight;
            }
	}
	break;
    case 1:
	// each voter gets equal sum weight of 1.0
	for ( i = 0; i < numv; i++ ) {
	    int j;
            counterweight = 0.0;
            for ( j = 0; j < numc; j++ ) {
                counterweight += fabs( they[i].getPref(j) );
            }
            counterweight = 1/counterweight;
            for ( j = 0; j < numc; j++ ) {
                talley[j] += they[i].getPref(j) * counterweight;
            }
	}
	break;
    case 0:
	// raw, same as maxHappiness
	for ( i = 0; i < numv; i++ ) {
	    int j;
	    for ( j = 0; j < numc; j++ ) {
		talley[j] += they[i].getPref(j);
	    }
	}
	break;
    }
    // find winner
    {
        double m = talley[0];
        winner = 0;
        for ( i = 1; i < numc; i++ ) {
            if ( talley[i] > m ) {
                m = talley[i];
                winner = i;
            }
        }
    }
    delete [] talley;
    if ( winnerR ) *winnerR = winner;
//    return pickOneHappiness( they, numv, winner );
}
